'''
****************************************************************************
*                                                                          *
*        BINARY SEARCH TREE module                                         *
*                                                                          *
*        supporting BE5B33PGE course at FEE CTU, Prague                    *
*                                                                          *
*                                                                          *
*        By Marko Berezovsky  (berezovs@fel.cvut.cz)     May 2024          *
*                                                                          *
****************************************************************************
'''

import random

# ============================================================================
#                class       N O D E
# ============================================================================

class Node:
    """A node in a Binary Search Tree (BST).
    Node in this implementation does NOT contain reference 
    to its parent. The goal of the implementation is to keep 
    the node structure and the BST operations maximally simple.
    
    Node attributes:
    ---------------- 
        left  : Node    Reference to the left child node.     
        right : Node    Reference to the right child node.
        key   : int     Key value, always unique in the whole BST. 
        value : int     Numerical value stored in the node.
        
    Node methods:
    -------------
        __init__( aKey, aValue )   
                Node constructor. 
            
    """

    def __init__(self, aKey, aValue ):
        """
        Node constructor. It creates a new node with key and value
        equal to aKey, aValue, respectively.
        """
        
        self.left = None     # mandatory structural property
        self.right = None    # mandatory structural property
        self.key = aKey     
        self.value = aValue
        
    #  -- End of class Node --------------------------------------------------
        
        
        
        

# ============================================================================
#                class       B I N A R Y    T R E E  
# ============================================================================
class BinarySearchTree:
    """
    A simple Binary Search Tree (BST) implementation.  
    The class is intended for BST demonstration, introductory experiments
    and various training applications. 
    The class contains only a single attribute, the reference
    to the root node. 
    No other information (size, depth, etc.) about the tree is kept.
    The class provides standard basic BST methods: find, insert, delete,
    and a few more methods suitable for demonstration and experimentation.
    
    
    attributes
    ---------- 
        root  : Node     Reference to the BST root node.     
    
    
    public methods
    --------------
        __init__( )   
                Constructor, creates  a new empty BST, root is set to None.
        
        find( aKey )
                Standard BST find method, returns reference to the node
                in which node.key == aKey. Returns None if such node
                does not exist in the tree.
        
        insert( aKey, aValue )       
                Standard BST insert method. A new node 
                Node( aKey, aValue ) is inserted in the tree if
                aKey is not present in BST yet. Otherwise the method
                has no effect.
      
        delete( aKey )
                Standard BST delete method. A node with node.key == aKey
                is deleted from BST. If such node is not in BST, the method
                has no effect.

        isEmpty( )
                True iff root == None.
                        
        nNodes( node )
                Number of nodes in the subtree rooted in node. 
                Number of nodes in BST when node is the BST root.

        depth( node )
                Depth of the subtree rooted in node. 
                Depth of BST when node is the BST root.
               
                
        randomBST( N, keyMin, keyMax, valueMin, valueMax, aSeed )
                Builds a BST populated with N nodes, node attributes
                are selected randomly in the given ranges.
         
        balance( )
                Reorganizes BST structure so that the depth of BST
                is minimized. Runs in Theta( N log N ) time, when
                BST is balanced, otherwise in O( N^2 ) time.
        
        display( )
                Displays BST in the text output. 
                
  
    Implementation Remarks 
    ----------------------
        All methods are implemented recursively. In Python versions 
        with limited recursion depth, it may be sometimes useful to 
        increase the recursion depth limit appropriately 
             import sys
             sys.setrecursionlimit( <yourOwnChoiceOfLimit> )
        When a monotone sequence of keys is inserted, the depth of BST, 
        and consequently the depth of recursion, may easily grow fast.
        However, if the tree is built randomly, e.g. by the provided 
        buildRandom function, the probability of reaching any 
        recursion limit is negligible.
       
    """


    def __init__( self ):
        """Construct a BST with no nodes. Tree root is None."""
        self.root = None




    def isEmpty( self ):
        """Returns True iff the BST contains no node."""
        return self.root == None
      



    def nNodes( self, node ):
        """
        Returns the number of nodes in the subtree rooted in node. 
        Returns the number of nodes in BST when node is the BST root.
        """
        if node == None: return 0
        return 1 + self.nNodes( node. left ) + self.nNodes( node.right )
  

  
    
    def depth( self, node ):
        """
        Returns the depth of the subtree rooted in node. 
        Returns the depth of BST when node is the BST root.
        """        
        if node == None: return -1
        return 1 + max( self.depth( node.left ), self.depth( node.right)  )

  

 
    def find( self, aKey ):
        """Standard BST find operation.
        
        Parameters
        ----------
            aKey : int     A key to be found in BST
            
        Returns
        -------
            Node           A reference to the node where node.key == aKey. 
                           Returns None when such node is not in BST.                  
        """
        return self._find( self.root, aKey )
      

    def _find( self, node, aKey ):
        # returns reference to the node with the key, 
        # or None if the key is not in the tree 
        
        if node == None: return None
      
        # search in L or R subtree? 
        if aKey < node.Key: return self._find( node.left )
        if node.Key < aKey: return self._find( node.right )
      
        # aKey == node.key
        return node
      

        

    def insert( self, aKey, aValue ):
        """Standard BST insert operation. 
        
        Parameters
        ----------
            aKey   : int   Key of the inserted node.     
            aValue : int   Value of the inserted node.
            
        Returns
        -------
            boolean        False when a node with node.key == aKey already
                           exists in the BST and a new node was not inserted.
                           True otherwise. 
        """
        self.root, didIt = self._insert( self.root, aKey, aValue )
        return didIt
        
        
    def _insert( self, node, aKey, aValue ):
        
        # place for new node found 
        if node == None: return Node( aKey, aValue ), True  

        # avoid key duplicates in BST
        if node.key == aKey: return node, False    # make no change 
        
        # otherwise continue in left or right subtree
        if aKey < node.key: 
            node.left, didIt = self._insert( node.left,  aKey, aValue )
        else:               
            node.right, didIt = self._insert( node.right, aKey, aValue )   
        return node, didIt




    
    def randomBST( self, N, keyMin, keyMax, valueMin, valueMax, aSeed ):
        """Creates a BST by inserting nodes with keys and values chosen from 
        uniformly random distribution in a given range.
        The expected depth of BST is Theta(log N),                
        the expected build time is therefore Theta( N log N ). 
        The BST is balanced, albeit not perfectly.
        
        Parameters
        ----------
            N        : int        
            keyMin   : int    Keys range in BST is (keyMIn, KeyMax).
            keyMax   : int    Keys range in BST is (keyMIn, KeyMax). 
            valueMin : int    Values range in BST is (valueMIn, valueyMax).
            valueMax : int    Values range in BST is (valueMIn, valueyMax).
            aSeed    : int    Random number generator seed.  
            
        Returns
        -------
            boolean        False when N > keyMax-keyMin and a BST cannot be 
                           constructed. In such case, no change in current BST 
                           occurs. True otherwise.
        """  
        if N > keyMax-keyMin: return  False
      
        random.seed( aSeed )
        keys = list( range( keyMin, keyMax) )
        random.shuffle(keys)
        
        self.root = None; # be harsh, just destroy any previous tree here
         
        for k in range( N ):
            value = random.randrange( valueMin, valueMax )
            self.insert( keys[k], value )
        return True  




    def delete( self, aKey ):
        """Standard BST delete operation. 
        
        Parameters
        ----------
            aKey : int     A key in the node which is to be deleted.      
            
        Returns
        -------
            boolean        True when a node with node.key == aKey was found 
                           and deleted in BST. False otherwise. 
        """
        self.root, didIt = self._delete( self.root, aKey )
        return didIt
        
    
    def _delete( self, node, aKey ):
        
        if node == None:  return None, False

        # either 1. delete in L or R subtree 
        if aKey < node.key:   
           node.left, didIt  = self._delete( node.left, aKey )
           return node, didIt
        
        if node.key < aKey:
           node.right, didIt = self._delete( node.right, aKey )
           return node, didIt
        
        # or    2. delete this node
        
        #  2.a   case of one child or no children
        if node.left  == None:  return node.right, True
        if node.right == None:  return node.left, True

        #  2.b   case of both children 
        # find substitute and store its data here
        leftmostInRsubtree =  node.right
        while leftmostInRsubtree.left != None: 
              leftmostInRsubtree = leftmostInRsubtree.left
        node.key = leftmostInRsubtree.key
        node.value = leftmostInRsubtree.value
        
        # recurse to delete the substitute
        node.right, didIt = self._delete( node.right, node.key )
        return node, didIt
    


       
    def balance( self ):
        """Reorganizes BST structure so that the depth of BST is minimized.
        No parameters and no return value. 
        
        Remarks
        -------
            The Left-right order of nodes in the tree is preserved.
            That is, as before, all keys in the left subtree of a node 
            are smaller than this node key, and all keys in the right subtree 
            of a node are bigger than this node key. Also, technically 
            speaking, the inorder traversal of the old and the new tree
            results in the same sequence of visited nodes.
            The sizes of the left and the right subtree of any node
            are either equal, or the left subtree size is bigger by 1
            then the right subtree size.
            Method runs in Theta( N log N ) time, when
            BST is balanced, otherwise in O( N^2 ) time.
        """
        nodesList = []
        self._nodesToListRecur( self.root, nodesList )
        self.root = self._balance( 0, len(nodesList), nodesList )


    def _nodesToListRecur( self, node, nodesList ):
        if node == None: return
      
        self._nodesToListRecur( node.left, nodesList )
        nodesList.append( node ) 
        self._nodesToListRecur( node.right, nodesList )
        return
        

    def _balance( self, iFrom, iTo, nodesList ):
       
          # empty range of nodes
        if iTo-iFrom <= 0: return None
        
        # local root is in the middle of the current range 
        iMid = (iFrom+iTo) // 2
        localRoot = nodesList[iMid]
        
        # connect local root to its L and R children
        localRoot.left  = self._balance( iFrom, iMid, nodesList )
        localRoot.right = self._balance( iMid+1, iTo, nodesList )
        
        return localRoot
  



    def display( self ):
        """Displays BST on standard text output. To keep the displayed
        width of the tree minimal, a node occupies two lines, the key
        is above the value.
        No input parametrs.

        Returns
        -------
            int        Width of BST in the text output expressed in
                       number of text columns.
        """
        if self.root == None:
           print( 'Empty tree' )
           return

        treeDepth = self.depth( self.root )
        canvas = [ '' for i in range( 2*treeDepth+2 ) ]
        treeWidth, foo, foo = self._putAtCoord( self.root, 0, 0, canvas )

        for row in canvas:
                print( row )

        print( '-' * treeWidth, '|nodes| =', self.nNodes( self.root ) )

        return treeWidth


    def _equalizeStrLen( self, s1, s2 ):
        # shorter string is aligned to the right by prepending some spaces to it
        if len(s1) < len(s2) : s1 = (' ' * (len(s2)-len(s1)) ) + s1
        else:                  s2 = (' ' * (len(s1)-len(s2)) ) + s2
        return s1, s2


    def _putAtCoord( self, node, LtopWidth, depth, canvas ):
        # 1. returns the total width ( = first unoccupied x coord )
        # of all nodes to the left of the current node
        # and of the nodes in the right subtree of the node
        # LARS = Left All, Right Subtree
        #
        # 2. returns the first and the last+1 x coord of the
        # current node in the canvas
        #
        # The width of the node is the number of characters
        # in the longer of string representation oof its key and value

        if node == None: return LtopWidth, LtopWidth, LtopWidth

        # calculations
        LARS_L, Lx1, Lx2 = self._putAtCoord( node.left, LtopWidth, depth+1, canvas )

        keyStr, valueStr = self._equalizeStrLen( str(node.key), str(node.value) )
        nodeLen = len( keyStr )

        LARS_R, Rx1, Rx2 = self._putAtCoord( node.right, LARS_L + nodeLen, depth+1, canvas )

        canvYcoord = depth*2 # because node occupies two lines
        
        # drawing
        # spaces before L child
        canvas[canvYcoord]   +=  ' ' * (Lx1 - len(canvas[canvYcoord]))
        canvas[canvYcoord+1] +=  ' ' * (Lx1 - len(canvas[canvYcoord+1]))
        # edge to L child
        canvas[canvYcoord]   +=  ' ' * (LARS_L - len(canvas[canvYcoord]))  
        canvas[canvYcoord+1] +=  '_' * (LARS_L - len(canvas[canvYcoord+1]))  
        # node itself
        canvas[canvYcoord]   += keyStr
        canvas[canvYcoord+1] += valueStr
        
        # edge to R child
        canvas[canvYcoord]   +=  ' ' * (Rx2 - len(canvas[canvYcoord]))
        canvas[canvYcoord+1] +=  '_' * (Rx2 - len(canvas[canvYcoord+1]))
        
        # return values
        nodex1 = LARS_L
        nodex2 = nodex1 + nodeLen
        return LARS_R, nodex1, nodex2  


    #  -- End of class BinarySearchTree --------------------------------------



if __name__ == "__main__" :

  print( ' This a simple Binary Search Tree module.' )
  print( ' See usage examples in bstexample*.py files. ' )


